#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <strstream.h>

#define MAXREALDIM 20   // you are free to modify this
#define DEFINED_REAL_SPEED_CONSTANTS 14
#define MAX_LEVEL   8

//  Radii are rounded in binary domains. M*threshold is not rounded.
//  That's why the output slightly differs from the values presented
//  earlier.

signed char real=1;

int         max_level=MAX_LEVEL;

long int    N=10000,    // evaluation number
			dim=1;

double      diam=10.0,
			M=0.0,      // max species number
			V[DEFINED_REAL_SPEED_CONSTANTS]={
				 0.0, .25, M_2_PI/3.0, 3.0/16.0, 4*M_2_PI/15.0, 10.0/64.0,
				 8*M_2_PI/35.0, 35.0/256.0, 64*M_2_PI/315.0,
				 63.0/512.0, 128*M_2_PI/693.0, 231.0/2048.0,
				 512*M_2_PI/3003.0, 858.0/8192.0 };

// -------------------------------------------------------------------------

double V_(long int dim) {
//
//  Using that binom(x,(x-1)/2)==.5*binom(x+1,(x+1)/2) and
//  binom(2n,n) is approx. 2**(2n)/sqrt(pi*n)
//
	return 1.0/sqrt( M_PI * (2 * (dim+1)) );
};

double v(double x) {
  if(real) return (dim<DEFINED_REAL_SPEED_CONSTANTS ? V[dim]*x : V_(dim)*x);
  else return 3.0/11.0*sqrt(x);
};

// -------------------------------------------------------------------------

double B_r(double x) {
  double s=0,z=x;
  for(int i=1; i<=max_level; ++i,z*=x) s+=1.0/v(diam*z);
  return s-(double)N/(double)diam/M;
};

// -------------------------------------------------------------------------

int main(int argc,char** argv) {

  double   a=0.0,
           b=1.0,
           beta,
           k=1.0,B,K,C,
           X[MAX_LEVEL+1],    // it. num. for levles 1 to MAX_LEVEL (x[0]=0)
           R[MAX_LEVEL+1];    // radius function

  int      i,j,
		   pop_size=100;

  long int Xr[MAX_LEVEL+1],  // rounded iteration number
           new_cost,   // cost of evolution after new_species
           stab_cost,  // cost of evolution after stabilize
           max_sp_n;

  // ---- reading parameters ------------------------------------------------

  if(argc==1) {
    cout
<< "Creates a GAS file for the given parameters.                          \n"
<< "The result is directed to the standard output.                      \n\n"
<< "beta [-bDIM|-rDIM|-nEVALS|-dDIAM|-tTHRESHOLD|-pPOPSIZE|-mMAXSPEC|  \n"
<< "      -lMAXLEVEL]                                                \n\n"
<< "        -b/-r    : binary/real domains                                \n"
<< "        DIM      : dimension of domain (max " << MAXREALDIM << " if real)"
<< "\n        EVALS    : number of function evaluations                   \n"
<< "        DIAM     : diameter of domain                                 \n"
<< "        THRESHOLD: from interval (0,1]                                \n"
<< "        POPSIZE  : population size                                    \n"
<< "        MAXSPEC  : maximal allowed number of species                  \n"
<< "        MAXLEVEL : max value of strict (from [1," << MAX_LEVEL << "])\n\n"
<< "defaults : [ -r1 -n10000 -d10.0 -t1 -p100 -m[POPSIZE/4] -l" <<
   MAX_LEVEL << " ]\n";
   return 0; };

  for(i=1; i<argc; ++i) {
    j=0;
    if(argv[i][0]=='-') ++j;
    switch(argv[i][j]) {
      case 'r' :
        if(argv[i][j+1]!=0)
          if((istrstream(argv[i]+j+1) >> dim).fail() ||
             dim < 1 || dim > MAXREALDIM ) {
            cerr << "Bad dimension. (" << argv[i] << ")\n"; exit(1); };
		real=1;
        break;
      case 'b' :
        if(argv[i][j+1]!=0)
          if((istrstream(argv[i]+j+1) >> dim).fail() ||
             dim < 1) {
            cerr << "Bad dimension. (" << argv[i] << ")\n"; exit(1); };
        diam=dim;
        real=0;
        break;
      case 'd' :
        if((istrstream(argv[i]+j+1) >> diam).fail() ||
		   diam <= 0) {
           cerr << "Bad diameter. (" << argv[i] << ")\n"; exit(1); };
        if(!real) dim=(long int)(diam=floor(diam));
        break;
      case 'n' :
        if((istrstream(argv[i]+j+1) >> N).fail() ||
           N < 1) {
          cerr << "Bad iteration number. (" << argv[i] << ")\n"; exit(1); };
        break;
      case 'm' :
        if((istrstream(argv[i]+j+1) >> M).fail() ||
           M < 1) {
		  cerr << "Bad max. spec. number. (" << argv[i] << ")\n"; exit(1); };
        break;
      case 't' :
        if((istrstream(argv[i]+j+1) >> k).fail() ||
           k <= 0 || k > 1) {
          cerr << "Bad threshold. (" << argv[i] << ")\n"; exit(1); };
        break;
      case 'p' :
        if((istrstream(argv[i]+j+1) >> pop_size).fail() ||
           pop_size < 2 ) {
          cerr << "Bad population size. (" << argv[i] << ")\n"; exit(1); };
        break;
	  case 'l' :
        if((istrstream(argv[i]+j+1) >> max_level).fail() ||
           max_level < 1 || max_level > MAX_LEVEL ) {
          cerr << "Bad max. level. (" << argv[i] << ")\n"; exit(1); };
        break;
      default :
        cerr << "Unknown switch: " << argv[i] << "\n";
        break;
    };
  };
  if(M==0.0) {
    max_sp_n=pop_size/4;
	M=floor(pop_size/4.0)*k;
  }
  else {
    max_sp_n=(long int)M;
    M*=k;
  };

  // ---- checking feasibility condition ------------------------------------

  if(M > pop_size) {
    cerr << "Population size should be greater than max. spec. number.\n";
    exit(1); };
  stab_cost=(int)(floor((pop_size*.1+1)/2)*2)*15+pop_size/2;
  new_cost=(int)(floor((pop_size*.1+1)/2)*3)*15+pop_size/2;
  if( (N-=max_level*new_cost) <= 0 ||
       N/(diam*M) <= max_level/v(diam)) {
    cerr << "No solution exists for the given parameter setting.\n";
    exit(1);
  };

  // ---- approximating beta ------------------------------------------------

  B=B_r(b);
  while(fabs(b-a)>1e-10) {
    k=(a+b)/2;
    K=B_r(k);
    if(B*K<0.0) a=k;
    else { b=k; B=K; };
  };
  beta=b;

  // ---- determining R, X and Xr -------------------------------------------

  R[0]=diam;
  X[0]=0.0;
  k=0;
  for(i=1; i<=max_level; ++i) {
    R[i]=R[i-1]*beta;
    k+=1/v(R[i]);
  };
  if(!real) { // rounded radii for binary domains
    k=0;
    for(i=1; i<=max_level; ++i) {
      R[i]=floor(R[i]+.5);
      if(R[i]==0) R[i]=1;
      k+=1/v(R[i]);
    };
  };
  X[0]=0.0;
  for(i=1; i<=max_level; ++i) {
    X[i]=N/k/v(R[i]);
    X[0]+=X[i];
  };
  Xr[0]=0;
  for(i=1; i<=max_level; ++i) {
    Xr[i]=(long int)(floor( (X[i]+stab_cost/2.0)/stab_cost ));
    Xr[0]+=Xr[i];
  };

  // ---- creating GAS file to standard out ---------------------------------

  cout << "// automatically generated GAS file\n\n" <<
       "// C = " << (double)N/k << "\n" <<
       "// max cost of evolution after new_species : " << new_cost << "\n" <<
       "// cost of evolution after stabilize :       " << stab_cost << "\n" <<
       "// all cost after roundings:                 " <<
       new_cost*max_level+stab_cost*Xr[0] << "\n" <<
       "// costs before roundings:\n";
  for(i=1; i<=max_level; ++i) cout << "// X[" << i << "] = " << X[i] << "\n";
  cout << "\ntask_parameter ?? < ?? >\n\n" <<
       "max_species_number " << max_sp_n <<
       "\n\nspecies_radius <\n";
  for(i=0; i<=max_level; ++i) cout << "  " << R[i] << "\n";
  if(i!=MAX_LEVEL+1) cout << 
    "// The following radius values are needed because of the syntax.\n";
  for(; i<=MAX_LEVEL; ++i) cout << "  " << R[max_level] << "\n";
  cout << ">\n\n" <<
       "create_population " << pop_size << "\n\n" <<
       "program {\n" <<
       "randomize\n";
  for(i=1; i<=max_level; ++i) cout <<
      "\nstrict " << i << "\n\n" <<
      "new_species\n" <<
      "repeat { evolution } 1 times\n"
      "//print species_number\n" <<
      "stabilize\n" <<
      "repeat { evolution } " << Xr[i] << " times\n" <<
      "//print species_number\n";
  cout << "\nsave 3 stdout\n}\n";
  return 0;
};

