#include <fstream.h>
#include "gainfo.h"
#include "compiler.h"

#define BAD_PAR {cerr << "Bad parameter in line " << line << ".\n"; exit(1);}

GA*      ga=NULL;
GAinfo*  gainfo;
task*    Task=NULL;
double*  task_par;
double   spec_rad[9];
int      start_pop_size,
	 line=1,
         max_sp_num=-1;


//-------------------------------------------------------------------------

void run(COMMAND* prog) {
  while(prog != NULL) { prog->action(); prog=prog->next; };
};

//-------------------------------------------------------------------------

void REPEAT::action() {
  int i;
  for(i=0; i<iter; ++i) run(program);
};

void RELAX::action() {
  ga->dying_off_relaxation(int(1000*(1+level)));
  ga->selection_relaxation(int(1000*(1+level)));
  ga->dying_off_error_(int((1+level)*(start_pop_size/2)));
};

void COMPETE::action() {
  ga->dying_off_relaxation(int(1000*(1-level)));
  ga->selection_relaxation(int(1000*(1-level)));
  ga->dying_off_error_(int((1-level)*(start_pop_size/2)));
};

void STABILIZE::action() {
  ga->freeze();
  ga->dying_off_error_(2);
  ga->dying_off_relaxation(10000);
  ga->selection_relaxation(0);
};

void NEW_SPECIES::action() {
  ga->thaw();
  ga->dying_off_error_(int(start_pop_size*0.8));
  ga->dying_off_relaxation(1000);
  ga->selection_relaxation(1000);
};

void EVOLUTION::action() {
  int i;
  for(i=0; i<10; ++i) ga->activity();
  ga->immigration(1.5);
  for(i=0; i<5; ++i) ga->activity();
};

void SAVE::action() {
  ofstream of;
  if(name[0]!=-1) {
    of.open((char*)name,ios::app);
    if(!of) {
      cerr << "Unable to open file : " << name << "!\n"; exit(1); };
    gainfo->save(*ga,of,mode);
  }
  else
    gainfo->save(*ga,cout,mode);
  of.flush();
};

void STRICT::action() { ga->strict(level); };

void STRICT_INC::action() { ga->strict( -1 ); };

void STRICT_DEC::action() { ga->strict( -2 ); };

void FREEZE::action() { ga->freeze(); };

void THAW::action() { ga->thaw(); };

void _DYING_OFF_RELAXATION::action() { ga->dying_off_relaxation(level); };

void _SELECTION_RELAXATION::action() { ga->selection_relaxation(level); };

void _DYING_OFF_ERROR::action() { ga->dying_off_error_(level); };

void _IMMIGRATION::action() { ga->immigration(growth); };

void _BASE_ACTIVITY::action() { ga->activity(); };

void _P_MUTATION::action() { ga->P_m_(mut); };

void _POPULATION_GROWTH::action() { ga->dpop_(growth); };

void RANDOMIZE::action() { RaNdOmIzE; };

void PRINT::action() {
  if(text[0]==-1)
    cout << "The number of the species : " << gainfo->species_number(*ga)
         << "\n";
  else cout << text << "\n";
};

//-------------------------------------------------------------------------

int nexttoken(fstream& f) {
  int         i,j;
  signed char token_list[TOKEN_NUM],
              c,
              empty;

  while(!f.eof()) {
    f.get(c);
    if(c=='/') {
      if(f.get() != '/') return -4;   // bad remark ?
      while(!f.eof() && f.get()!='\n');
      ++line; }
    else if(c!=' ' && c!='\n' && c!='\t') break;
	 else if(c=='\n') ++line;
  };
  if(f.eof()) return -2;              // no more data

  for(i=0; i<TOKEN_NUM; ++i) token_list[i]=1;   // every token is possible

  for(j=0; 1; ++j) {
    if(c=='/') {
      if(f.get() != '/') return -4;   // bad remark ?
      while(!f.eof() && f.get()!='\n');
      ++line;
      if(f.eof()) return -3;          // unexpected eof
      else { f.get(c); --j; continue; }
    };
    for(i=0; i<TOKEN_NUM; ++i)
      if(token_list[i])
	if(X[i][j] !=c ) token_list[i]=0;
	else
	  if(X[i][j+1]==0) return i;  // that's it
    for(empty=1,i=0; i<TOKEN_NUM && empty; ++i)
      if(token_list[i]) empty=0;
    if(empty) return -1;              // there is an error in the source file
    if(f.eof()) return -3;            // unexpected eof
    f.get(c);
    if(c == '\n') ++line;
  };
};

//-------------------------------------------------------------------------

void read_nothing(fstream& f) {
  signed char c=f.peek();
  while(c==' ' || c=='\t' || c=='\n' || c=='/') {
    f.get(c);
    if(c=='\n') ++line;
    if(c=='/')
      if(f.peek()=='/') {
        while(!f.eof() && f.get()!='\n');
        ++line; }
      else BAD_PAR;
    c=f.peek();
  };
};

//-------------------------------------------------------------------------

int NUM1(fstream& f) {
  int i;
  read_nothing(f);    // to keep variable 'line' correct
  f >> i;
  if(f.fail() || i<0) BAD_PAR;
  return i;
};

//-------------------------------------------------------------------------

int NUM2(fstream& f) {
  int i;
  read_nothing(f);    // to keep variable 'line' correct
  f >> i;
  if(f.fail() || i<=0) BAD_PAR;
  return i;
};

//-------------------------------------------------------------------------

int NUM3(fstream& f) {
  int i;
  read_nothing(f);    // to keep variable 'line' correct
  f >> i;
  if(f.fail() || i<0 || i>8) BAD_PAR;
  return i;
};

//-------------------------------------------------------------------------

int NUM4(fstream& f) {
  int i;
  read_nothing(f);    // to keep variable 'line' correct
  f >> i;
  if(f.fail() || i<0 || i>3) BAD_PAR;
  return i;
};

//-------------------------------------------------------------------------

double REAL(fstream& f) {
  double d;
  read_nothing(f);    // to keep variable 'line' correct
  f >> d;
  if(f.fail()) BAD_PAR;
  return d;
};

//-------------------------------------------------------------------------

double REAL1(fstream& f) {
  double d;
  read_nothing(f);    // to keep variable 'line' correct
  f >> d;
  if(f.fail() || d<=0) BAD_PAR;
  return d;
};

//-------------------------------------------------------------------------

double REAL2(fstream& f) {
  double d;
  read_nothing(f);    // to keep variable 'line' correct
  f >> d;
  if(f.fail() || d<=1) BAD_PAR;
  return d;
};

//-------------------------------------------------------------------------

double REAL3(fstream& f) {
  double d;
  read_nothing(f);    // to keep variable 'line' correct
  f >> d;
  if(f.fail() || d<0 || d>1) BAD_PAR;
  return d;
};

//-------------------------------------------------------------------------

void TASK_PAR(fstream& f) {
  int i=NUM1(f);

  if(nexttoken(f) != 2) {
    cerr << "Waited for '" << X[2] << "' in line " << line << ".\n";
    exit(1); };

  if((task_par=new double[i+1])==NULL) fewmem();

  for(int j=0; j<i; ++j) task_par[j]=REAL(f);

  if(nexttoken(f) != 3) {
    cerr << "Waited for '" << X[3] << "' in line " << line << ".\n";
    exit(1); };

  if((Task=new task(task_par))==NULL) fewmem();
};

//-------------------------------------------------------------------------

void SPEC_RAD(fstream& f) {
  if(nexttoken(f) != 2) {
    cerr << "Waited for '" << X[2] << "' in line " << line << ".\n";
    exit(1); };

  for(int j=0; j<9; ++j) spec_rad[j]=REAL1(f);

  if(nexttoken(f) != 3) {
    cerr << "Waited for '" << X[3] << "' in line " << line << ".\n";
    exit(1); };
};

//-------------------------------------------------------------------------

void CREATE_POP(fstream& f) {
  int i=NUM2(f);
  if(i==1) {
    cerr << "Population size must be at least 2 in line " << line << " !\n";
    exit(1); };

  if(Task==NULL) // task_par is NULL here
    if((Task=new task(task_par))==NULL) fewmem();

  if(max_sp_num>i) max_sp_num=i-1;
  if(max_sp_num==-1) {
    if((ga=new GA(*Task,i,spec_rad))==NULL) fewmem();
  }
  else {
    if((ga=new GA(*Task,i,spec_rad,(double)max_sp_num/i))==NULL) fewmem();
  };
  start_pop_size=i;
};

//-------------------------------------------------------------------------

void STRING(fstream& f, signed char *name) {
  signed char c;
  int         j;

  for(j=0,f.get(c); !f.eof() && c!='"' && j<254; f.get(c),++j) name[j]=c;
  if(c != '"') {
    cerr << "Bad ""string"" in line " << line << ".\n";
    exit(1); };
  name[j]=0;
};

//-------------------------------------------------------------------------

COMMAND* PROG(fstream&);

void COMM(fstream& f, int i, COMMAND*& co) {
  signed char name[255];
  COMMAND*    p;
  ostream*    of;

  switch(i) {

    case 8: // repeat
      if(nexttoken(f) != 0) { // {
	cerr << "Waited for '" << X[0] << "' in line " << line << ".\n";
	exit(1); };
      p=PROG(f);
      if((co=new REPEAT( p, NUM1(f))) == NULL) fewmem();
      if(nexttoken(f) != 9) { // times
	cerr << "Waited for '" << X[9] << "' in line " << line << ".\n";
	exit(1); };
      break;

    case 10:
      if((co=new RELAX( REAL3(f))) == NULL) fewmem();
      break;

    case 11:
      if((co=new COMPETE( REAL3(f))) == NULL) fewmem();
      break;

    case 12:
      if((co=new STABILIZE) == NULL) fewmem();
      break;

    case 13:
      if((co=new EVOLUTION) == NULL) fewmem();
      break;

    case 14: // save
      i=NUM4(f);
      switch(nexttoken(f)) {
        case 27: // "
          STRING(f,name);
          break;
        case 30: // stdout
          name[0]=-1;name[1]=0;
          break;
        default:
	  cerr << "Waited for '" << X[30] << "' or '"
               << X[27] << "' in line " << line << ".\n";
	  exit(1); break;
      };
      if((co=new SAVE( i, name)) == NULL) fewmem();
      break;

    case 15:
      if((co=new STRICT( NUM3(f))) == NULL) fewmem();
      break;

    case 16: // ++strict
      if((co=new STRICT_INC) == NULL) fewmem();
      break;

    case 17:
      if((co=new STRICT_DEC) == NULL) fewmem();
      break;

    case 18:
      if((co=new FREEZE) == NULL) fewmem();
      break;

    case 19:
      if((co=new THAW) == NULL) fewmem();
      break;

    case 20:
      if((co=new _DYING_OFF_RELAXATION( NUM2(f))) == NULL) fewmem();
      break;

    case 21:
      if((co=new _SELECTION_RELAXATION( NUM2(f))) == NULL) fewmem();
      break;

    case 22:
      if((co=new _DYING_OFF_ERROR( NUM2(f))) == NULL) fewmem();
      break;

    case 23:
      if((co=new _IMMIGRATION( REAL2(f))) == NULL) fewmem();
      break;

    case 24:
      if((co=new _BASE_ACTIVITY) == NULL) fewmem();
      break;

    case 25:
      if((co=new _P_MUTATION( NUM2(f))) == NULL) fewmem();
      break;

    case 26:
      if((co=new _POPULATION_GROWTH( REAL2(f))) == NULL) fewmem();
      break;

    case 28:
      if((co=new RANDOMIZE) == NULL) fewmem();
      break;

    case 29: // print
      switch(nexttoken(f)) {
        case 27: // "
          STRING(f,name);
          break;
        case 31: // species_number
          name[0]=-1;
          name[1]=0;
          break;
        default:
          cerr << "Waited for a '" << X[27] << "' or '" <<
                  X[31] << "' in line " << line << ".\n";
          exit(1); break;
      };
      if((co=new PRINT(name)) == NULL) fewmem();
      break;

    case 32:
      if((co=new NEW_SPECIES) == NULL) fewmem();
      break;

    default:
      cerr << "Waited for a command in line " << line << ".\n"; exit(1);
  };
};


//-------------------------------------------------------------------------

COMMAND* PROG(fstream& f) {
  COMMAND *result=NULL,
	  *co;
  int      i=nexttoken(f);
  if(i<0) { cerr << "Syntax error in line " << line << ".\n"; exit(1); };

  if(i != 1) COMM(f,i,result);
  else return result;           // empty program-body

  co=result;
  while((i=nexttoken(f)) != 1) {
    if(i<0) { cerr << "Syntax error in line " << line << ".\n"; exit(1); };
    COMM(f,i,co->next);
    co=co->next;
  };
  return result;
};

//-------------------------------------------------------------------------

COMMAND* S(fstream& f) {
COMMAND* result;

  if(nexttoken(f)!=4) {
    cerr << "Waited for '" << X[4] << "' in line " << line << ".\n";
    exit(1); };
  TASK_PAR(f);
  switch(nexttoken(f)) {
    case 33:
      max_sp_num=NUM2(f);
      if(nexttoken(f) != 5) {
        cerr << "Waited for '" << X[5] << "' in line " << line << ".\n";
        exit(1); };
    case 5: break;
    default:
      cerr << "Waited for '" << X[5] << "' or '" << X[33] <<
      "' in line " << line << ".\n";
      exit(1); break;
  };
  SPEC_RAD(f);
  if(nexttoken(f) != 6) {
    cerr << "Waited for '" << X[6] << "' in line " << line << ".\n";
    exit(1); };
  CREATE_POP(f);
  if(nexttoken(f) != 7) {
    cerr << "Waited for '" << X[7] << "' in line " << line << ".\n";
    exit(1); };
  if(nexttoken(f) != 0) {
    cerr << "Waited for '" << X[0] << "' in line " << line << ".\n";
    exit(1); };
  return PROG(f);
};

//-------------------------------------------------------------------------

int main(int argc,char** argv) {
  signed char name[80],
	      suf[5]=".gas",
	      i,j;
  fstream     f;
  COMMAND*    Task_;

  if((gainfo=new GAinfo)==NULL) fewmem();

  if(argc<2) {
    cout << "The name of the GAS source file : ";
    cin >> name;
    f.open((char*)name,ios::in);
    if(!f) {
      for(i=0; name[i]!=0; ++i); for(j=0; j<5; ++j) name[i+j]=suf[j];
      f.open((char*)name,ios::in); };
  }
  else {
    f.open(argv[1],ios::in);
    if(!f) {
      for(i=0; argv[1][i]!=0; ++i);
      for(j=0; j<5; ++j) argv[1][i+j]=suf[j];
      f.open(argv[1],ios::in); };
  };

  if(!f) {
    cerr << "Unable to open file " << (argc<2 ? (char*)name : argv[1]) 
         << " .\n";
    exit(1); };
  cout << '\n';

  Task_=S(f);
  run(Task_);
  return 0;
};
