#ifndef COMPILER_H
#define COMPILER_H

#define TOKEN_NUM 34

// New tokens should be inserted after the end of the array X.
// None of the tokens is allowed to be a prefix of another.

const char *const X [TOKEN_NUM] = { 
       "{",                          //   0
       "}",                          //   1
       "<",                          //   2
       ">",                          //   3
       "task_parameter",             //   4
       "species_radius",             //   5
       "create_population",          //   6
       "program",                    //   7
       "repeat",                     //   8
       "times",                      //   9
       "relax",                      //  10  10 and 11 are "out of fashion"
       "compete",                    //  11
       "stabilize",                  //  12
       "evolution",                  //  13
       "save",                       //  14
       "strict",                     //  15
       "++strict",                   //  16
       "--strict",                   //  17
       "freeze",                     //  18
       "thaw",                       //  19
       "_dying_off_relaxation",      //  20
       "_selection_relaxation",      //  21
       "_dying_off_error",           //  22
       "_immigration",               //  23
       "_base_activity",             //  24
       "_P_mutation",                //  25
       "_population_growth",         //  26
       "\"",                         //  27
       "randomize",                  //  28
       "print",                      //  29
       "stdout",                     //  30
       "species_number",             //  31
       "new_species",                //  32
       "max_species_number"          //  33

};

// Don't forget separating the items with a comma (,) and updating TOKEN_NUM !

//-------------------------------------------------------------------------

class COMMAND {

public:
  COMMAND() : next(NULL) {;};
  ~COMMAND() { if(next != NULL) delete next; };
  virtual void action() = 0;
  COMMAND* next;
};

//-------------------------------------------------------------------------

class REPEAT : public COMMAND {

private:
  int      iter;
  COMMAND* program;

public:
  REPEAT(COMMAND* c,int i) : iter(i), program(c) {;};
  ~REPEAT() { if(program != NULL) delete program; };
  void action();
};

//-------------------------------------------------------------------------

class RELAX : public COMMAND {

private:
  double level;

public:
  RELAX(double d) : level(d) {;};
  void action();
};

//-------------------------------------------------------------------------

class COMPETE : public COMMAND {

private:
  double level;

public:
  COMPETE(double d) : level(d) {;};
  void action();
};

//-------------------------------------------------------------------------

class STABILIZE : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class EVOLUTION : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class SAVE : public COMMAND {

private:
  int         mode;
  signed char name[255];

public:
  SAVE(int,signed char *);
  void action();
};

SAVE::SAVE(int i,signed char *n) : mode(i) {
  int j=0;
  for(;j<254 && n[j]!=0;++j) name[j]=n[j];
  name[j]=0;
};

//-------------------------------------------------------------------------

class STRICT : public COMMAND {

private:
  int level;

public:
  STRICT(int i) : level(i) {;};
  void action();
};

//-------------------------------------------------------------------------

class STRICT_INC : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class STRICT_DEC : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class FREEZE : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class THAW : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class _DYING_OFF_RELAXATION : public COMMAND {

private:
  int level;

public:
  _DYING_OFF_RELAXATION(int i) : level(i) {;};
  void action();
};

//-------------------------------------------------------------------------

class _SELECTION_RELAXATION : public COMMAND {

private:
  int level;

public:
  _SELECTION_RELAXATION(int i) : level(i) {;};
  void action();
};

//-------------------------------------------------------------------------

class _DYING_OFF_ERROR : public COMMAND {

private:
  int level;

public:
  _DYING_OFF_ERROR(int i) : level(i) {;};
  void action();
};

//-------------------------------------------------------------------------

class _IMMIGRATION : public COMMAND {

private:
  double growth;

public:
  _IMMIGRATION(double d) : growth(d) {;};
  void action();
};

//-------------------------------------------------------------------------

class _BASE_ACTIVITY : public COMMAND {

public:
  _BASE_ACTIVITY() {;};
  void action();
};

//-------------------------------------------------------------------------

class _P_MUTATION : public COMMAND {

private:
  int mut;

public:
  _P_MUTATION(int i) : mut(i) {;};
  void action();
};

//-------------------------------------------------------------------------

class _POPULATION_GROWTH : public COMMAND {

private:
  double growth;

public:
  _POPULATION_GROWTH(double d) : growth(d) {;};
  void action();
};

//-------------------------------------------------------------------------

class RANDOMIZE : public COMMAND {

public:
  void action();
};

//-------------------------------------------------------------------------

class PRINT : public COMMAND {

private:
  signed char text[255];

public:
  PRINT(signed char *);
  void action();
};

PRINT::PRINT(signed char *n) {
  int j=0;
  for(;j<254 && n[j]!=0;++j) text[j]=n[j];
  text[j]=0;
};

//-------------------------------------------------------------------------

class NEW_SPECIES : public COMMAND {

public:
  void action();
};


#endif
