#include "ga.h"

#define P_CAN_GO_IN(x) fen-p_fen[x]<=spec_rad[r_index]
#define O_CAN_GO_IN(x) fen-o_fen[x]<=spec_rad[r_index]


//---------------------------------------------------------------------------

GA::GA(task& t,int popsize, double* spec_r, double maxsp, signed char rnd) :
pop(t,popsize,rnd,maxsp)
{
  signed char i;
  task_=&t;
  P_m=30;
  if(spec_r==NULL) {
    spec_rad[0]=10;                                   // default
    for(i=1;i<9;++i) spec_rad[i]=spec_rad[i-1]/10; }  //         spec_rad
  else
    for(i=0;i<9;++i) spec_rad[i]=spec_r[i];
  doptimum=spec_rad[0];
  starting_num=pop.pop_size();
  if(starting_num>9) dpop=1.1;
  else dpop=1.5;
  dying_off_error=2;
};

//---------------------------------------------------------------------------

GA::GA(task& t, const individual* ia, int popsize, double* spec_r,
       double maxsp) :
pop(t,popsize,0,maxsp,ia)
{
  signed char i;
  task_=&t;
  P_m=30;
  if(spec_r==NULL) {
    spec_rad[0]=10;                                   // default
    for(i=1;i<9;++i) spec_rad[i]=spec_rad[i-1]/10; }  //         spec_rad
  else
    for(i=0;i<9;++i) spec_rad[i]=spec_r[i];
  doptimum=spec_rad[0];
  starting_num=pop.pop_size();
  if(starting_num>9) dpop=1.1;
  else dpop=1.5;
  dying_off_error=2;
};

//---------------------------------------------------------------------------

void GA::create_offspring() {
  pop.offspring[0]=pop.parent(1);
  pop.offspring[1]=pop.parent(2);
  pop.offspring[0].crossover(pop.offspring[1],*task_);
  pop.offspring[0].mutate(P_m,*task_);
  pop.offspring[1].mutate(P_m,*task_);
};

//---------------------------------------------------------------------------

void GA::divide(fenotype* p_fen,fenotype* o_fen) {
// virt_tree_pntr points to parent's species!
  int         in=0,r_index;
  fenotype    fen;
  signed char more=1;

  // parents

  if(more=pop.to_firstson()) {   // = is OK !
    while(more && in<3) {
      fen=pop.center();
      r_index=(8>pop.rad_index() ? pop.rad_index() : 8);
      switch(in) {
	case 0:if(P_CAN_GO_IN(0)) { pop.put_in("s1");in++;break; };
	case 1:if(P_CAN_GO_IN(1)) { pop.put_in("s2");in+=2; };break;
	case 2:if(P_CAN_GO_IN(0)) { pop.put_in("s1");in++; };break; };
      more=pop.to_brother(); };
    pop.to_dad(); };
  if(in != 3) { // couldn't put both into existing species
    if(in != 1) pop.put_in("s1+");     // new
    if(in != 2) pop.put_in("s2+"); };  //   species

  // offspring

  in=0;
  more=1;
  if(more=pop.to_firstson()) {  //virt_tree_pntr points to spec! (= is OK!)
    while(more && in<3) {
      fen=pop.center();
      r_index=(8>pop.rad_index() ? pop.rad_index() : 8);
      switch(in) {
	case 0:if(O_CAN_GO_IN(0)) { pop.put_in("u1");in++; };
	case 1:if(O_CAN_GO_IN(1)) { pop.put_in("u2");in+=2; };break;
	case 2:if(O_CAN_GO_IN(0)) { pop.put_in("u1");in++; };break; };
      more=pop.to_brother(); };
    pop.to_dad(); };
  if(in != 3)  // couldn't put both into existing species
    put_in_offspr(o_fen,in);
};

//---------------------------------------------------------------------------

void GA::put_in_offspr(fenotype* o_fen, int in) {
// virt_tree_pntr points to parent's species!
  int           r_index;
  fenotype      fen;
  signed char   more=1;
  for(;in<3 && more;more=pop.to_dad()) { // proceeds upwards
    fen=pop.center();
    r_index=(8>pop.rad_index() ? pop.rad_index() : 8);
    switch(in) {
      case 0:if(O_CAN_GO_IN(0)) { pop.put_in("u1");in++; };
      case 1:if(O_CAN_GO_IN(1)) { pop.put_in("u2");in+=2; };break;
      case 2:if(O_CAN_GO_IN(0)) { pop.put_in("u1");in++; };break; };
  };
  if(in != 3) { // root is the current species!
    if(in!=1) pop.put_in("u1");
    if(in!=2) pop.put_in("u2"); };
};

//---------------------------------------------------------------------------

void GA::activity() {
  fenotype      p_fen[2],o_fen[2];
  fenotype      *btwn;
  individual    *between;
  int           level,i,maxpopcount=int(dpop*starting_num);

  while(pop.pop_size()<maxpopcount) {

    pop.parent_selection();
    pop.parent_spec();        // virt_tree_pntr points to parent's species!
    level=pop.parent_level(); // 7 if last, 0 if root
    create_offspring();
    p_fen[0]=pop.parent(1).fen();
    p_fen[1]=pop.parent(2).fen();
    o_fen[0]=pop.offspring[0].fen();
    o_fen[1]=pop.offspring[1].fen();

    // inserting
    // if a random individual which is between the two parents is
    // weeker than any of the parents, we divide

    if(!pop.freeze) {

      if((btwn=new fenotype(pop.parent(1).fen(),pop.parent(2).fen()))==NULL)
        fewmem();
      if((between=new individual(*btwn, *task_))==NULL) fewmem();

      if(pop.parent(1).fit() > between->fit())
        if(level!=7 && pop.distinct_parents() &&
                       pop.parent(2).fit() > between->fit())
          divide(p_fen, o_fen);
        else put_in_offspr( o_fen );
      else {
        put_in_offspr( o_fen );
        if(pop.parent(2).fit() < between->fit()) { // between is prety good!
          pop.parent_spec();        // OK since wasn't divide
          pop.offspring[0]=*between;
          o_fen[0]=between->fen();
          put_in_offspr(o_fen, 2);  // put in between!
        };
      };
    delete between;
    delete btwn;
    }
    else put_in_offspr( o_fen );
  };
  pop.dying_off(dying_off_error,pop.pop_size()-starting_num);
  pop.fusion(doptimum,
             double(pop.pop_size()-starting_num)/(maxpopcount-starting_num));
  // if dying_off is unsuccesful fusion absorbes very small species too
};

//---------------------------------------------------------------------------

void GA::immigration_() {
  fenotype fen,centerfen=pop.center();
  int      i;
  double   r=spec_rad[ 8>pop.rad_index() ? pop.rad_index() : 8 ];

  // through the children (if there are)

  if(pop.to_firstson()) {
    do immigration_(); while(pop.to_brother());
    pop.to_dad(); };

  // immigration

  i=(int)((dpop-1)*pop.count());
  do {
    fen=fenotype(centerfen,r); // inside a circle with center centerfen rad r
    pop.offspring[0]=individual(fen,*task_);
    pop.put_in("u1"); }
  while(--i > 0);
};

//---------------------------------------------------------------------------

void GA::immigration(double d) {
  unsigned char spec[8]={ 0,0,0,0,0,0,0,0 };
  double        assist=dpop;
  pop.set_virt_tree_pntr(spec);
  if(d>1) dpop=d;
  immigration_();
  pop.dying_off(dying_off_error,pop.pop_size()-starting_num);
  pop.fusion(doptimum,0.0);
  dpop=assist;
};

//---------------------------------------------------------------------------

void GA::strict(int i) {
// i=-1 -> increment
// i=-2 -> decrement
  if(i==-1) i=pop.min_rad_index+1;
  if(i==-2) i=pop.min_rad_index-1;
  doptimum=spec_rad[pop.min_rad_index=(i<0 ? 0 : (i>8 ? 8 : i))];
};

//===========================================================================

void change_task_par(double* param,GA& ga) {
  ga.task_->new_parameter(param);
  ga.pop.update();
};
