#ifndef INDIVIDUAL_H
#define INDIVIDUAL_H
#include <limits.h>
#include <time.h>
#include <stdlib.h>
#include <iostream.h>

//===========================================================================

#ifndef RAND_MAX
#define RAND_MAX    32767
#endif
#define RaNdOmIzE   srand((unsigned)time(NULL))
#define RaNdOm(num) ((int)(( (double)rand()*(num) )/(RAND_MAX+1.0)))

//===========================================================================

class   task;
class   genotype;
class   fenotype;
class   GA;
class   GAinfo;

#include "changeme.h"

typedef double fitnesst;

//===========================================================================

class task {
friend class GAinfo;

private:
  double*  parameter_;  // a chance to pass parameters to the task
                        //                  (eg upper or lower bound)
                        // stores only a pointer !
public:
  task(double*);
  void     new_parameter(double*);
  double   parameter(int i) { return parameter_[i]; };
  fitnesst fit(fenotype&);
  fenotype fen(genotype&);
  genotype invfen(fenotype&);
};

//===========================================================================

ostream& operator<<(ostream&,genotype&);

class genotype {
friend class task;
friend ostream& operator<<(ostream&,genotype&);

private:
  GENOTYPE gen;

public:

  genotype() {;};
  genotype(GENOTYPE);
  genotype(const genotype&);
  ~genotype();
  genotype&   operator=(const genotype&);
  signed char operator==(genotype&);
  void        rnd_init();               // randomly initiates genotype
  void        operator*(genotype&);     // crossower
  signed char operator*(unsigned int);  // mutation; parameter is eg P_m
};

//===========================================================================

ostream& operator<<(ostream&,fenotype&);

class fenotype {
friend class GAinfo;
friend class task;
friend ostream& operator<<(ostream&,fenotype&);
friend void     change_task_par(double*,GA&);

private:
  FENOTYPE fen;

public:

  fenotype() {;};
  fenotype(FENOTYPE);
  fenotype(const fenotype&);
  fenotype(fenotype&,double);             // inside the given sphere
  fenotype(fenotype&,fenotype&);          // on the given section
  ~fenotype();
  fenotype&   operator=(const fenotype&);
  signed char between(fenotype&,fenotype&);// term between; you can redefine it
  double      operator-(fenotype&);       // term distance;you can redefine it
};

//===========================================================================

class individual {
friend class GAinfo;

private:
  genotype  gen_;
  fenotype  fen_;
  fitnesst  fit_;

public:
  individual() {;};
  individual(task& t) {
	      gen_.rnd_init();fit_=t.fit(fen_=t.fen(gen_)); };
  individual(genotype& g,task& t) { gen_=g;fit_=t.fit(fen_=t.fen(g)); };
  individual(fenotype& f,task& t) {
	      gen_=t.invfen(f);fit_=t.fit(fen_=t.fen(gen_)); }; // this way !
  individual(const individual& i) {
              gen_=i.gen_; fen_=i.fen_; fit_=i.fit_; };
  void      init(task& t) { fit_=t.fit(fen_=t.fen(gen_)); };
  genotype  gen() const { return gen_; };
  fenotype  fen() const { return fen_; };
  fitnesst  fit() const { return fit_; };
  signed char operator==(individual& e) { return gen_==e.gen_; };
  void      crossover(individual& e,task& t) {
              gen_*e.gen_;e.init(t);fit_=t.fit(fen_=t.fen(gen_)); };
  void      mutate(unsigned int pm,task& t) {
              if(gen_*pm!=0) fit_=t.fit(fen_=t.fen(gen_)); };
};

#endif
