#ifndef INDIVIDUAL_H
#include "individl.h"
#endif

#ifndef POPULATION_H
#define POPULATION_H

struct taxon_chart_;

typedef struct ind_list_ {
	  struct ind_list_*    next;
	  individual           e;
	  struct taxon_chart_* spec;
	} ind_list_t;

typedef struct taxon_chart_ {
	  struct taxon_chart_* firstson;
	  struct taxon_chart_* brother;
	  struct taxon_chart_* dad;
	  ind_list_t*          indvdls;   // individuals of species
	  ind_list_t*          lastind;   // the last ind. of close species
	  fitnesst             fitsum;    // fitness sum of close species
	  fitnesst             fitlowb;   // fitness lower bound
	  ind_list_t*          maxe;      // strongest ind. of close species
	  int                  countall;  // count of wide species
	  int                  count;
	  int		       rad_index; // index of spec_rad of species
	} taxon_chart_t;

class population {
friend class GAinfo;

private:
  taxon_chart_t   taxon_chart;
  taxon_chart_t*  virt_tree_pntr; // used to present virtual taxon_chart
  ind_list_t*     iterator;
  taxon_chart_t*  iter_tree;      // species being iterated
  individual      max_ind;        // temporary
  ind_list_t*     parents[2];     // for storing two parents
  fitnesst        Gfitlowb;       // global fitness lower bound
  task            *task_;         // pointer to the member of parent-GA
  int            species_number;
  int            max_species_number;
  void           species(taxon_chart_t*,unsigned char *);
  void           insert(const individual*);
  taxon_chart_t* insert_to_new(const individual*);
  signed char    kill(ind_list_t*);
  taxon_chart_t* pointer(unsigned char *);                // NULL if fails
  void      del_tree(taxon_chart_t*);                     // helps destruktor
  void      update_(taxon_chart_t*);                       // helps update
  void      fusion(double,double,taxon_chart_t*);          // helps fusion
  void      unite_if_needed(taxon_chart_t*,double,double); // helps fusion
  void      unite(taxon_chart_t*,taxon_chart_t*);          // helps fusion
  signed char del_holes(taxon_chart_t*);                  // helps fusion
  void      maximum(taxon_chart_t*,signed char);          // helps dying_relax
  void      iterstart(taxon_chart_t* f) { iterator=(iter_tree=f)->indvdls; };
  signed char iter_all_end(ind_list_t*);
  void      fitinf(taxon_chart_t*,signed char=0);
                                // maximalize fitlowb-s of given wide species

public:
  population(task&, int, signed char, double, const individual* i=NULL);
  ~population();
  individual  offspring[2];  // the two offspring
  signed char freeze;        // no new species created if 1
  int         min_rad_index; // minimal value of rad_index of new species
  void        iterstart(unsigned char*);
  individual* iteradd();
  signed char iterend();
  signed char iter_all_end() { return iter_all_end(iterator); };
  signed char set_virt_tree_pntr(unsigned char* s)
             { return (virt_tree_pntr=pointer(s))==NULL ? 0 : 1; };
  void      get_virt_tree_pntr(unsigned char* s)
             { species(virt_tree_pntr,s); };
  signed char to_brother() { if(virt_tree_pntr->brother==NULL) return 0;
               else { virt_tree_pntr=virt_tree_pntr->brother;return 1; }; };
  signed char to_dad() { if(virt_tree_pntr->dad==NULL) return 0;
               else { virt_tree_pntr=virt_tree_pntr->dad;return 1; }; };
  signed char to_firstson() { if(virt_tree_pntr->firstson==NULL) return 0;
               else { virt_tree_pntr=virt_tree_pntr->firstson;return 1; }; };
  int       rad_index() { return
               virt_tree_pntr->rad_index>8 ? 8 : virt_tree_pntr->rad_index; };
  int       count() { return virt_tree_pntr->count; };
  fenotype  center() { return virt_tree_pntr->maxe->e.fen(); };
  void      parent_spec() { virt_tree_pntr=parents[0]->spec; };
  int       parent_level();
  signed char distinct_parents() { return (parents[0]!=parents[1]); };
  individual parent(int a) { return parents[a==1 ? 0 : 1]->e; };
  void      parent_selection();
  void      put_in(const char *);
  int       pop_size() { return taxon_chart.countall; };
  void      dying_off(unsigned int,int);
  void      update();        // updates all ind.s + fields of taxon_chart
  void      sel_relax(unsigned int,taxon_chart_t *f);
  void      dying_relax(unsigned int);
  void      fusion(double d1,double d2)
                            // fusion of species too close to each other
             { fusion(d1, d2, &taxon_chart); del_holes(&taxon_chart); };
};

void fewmem();

#endif